//////////////////////////////////////////////////////////////////////////////////
// Company: TEXAS INSTRUMENTS
// Engineer: 
// 
// Create Date: 11/28/2021 04:02:36 PM
// Design Name: Deserializer for ADS9818
// Module Name: deser_cmos
// Project Name: 
// Target Devices: 
// Tool Versions: 
// Description: 
// 
// Dependencies: 
// 
// Revision:
// Revision 0.01 - File Created
// 
//////////////////////////////////////////////////////////////////////////////////

/*
* Copyright (c) 2022 – 2023 Texas Instruments Incorporated
*
* All rights reserved not granted herein.
* Limited License. 
*
* Texas Instruments Incorporated grants a world-wide, royalty-free,
* non-exclusive license under copyrights and patents it now or hereafter
* owns or controls to make, have made, use, import, offer to sell and sell ("Utilize")
* this software subject to the terms herein.  With respect to the foregoing patent
* license, such license is granted  solely to the extent that any such patent is necessary
* to Utilize the software alone.  The patent license shall not apply to any combinations which
* include this software, other than combinations with devices manufactured by or for TI (“TI Devices”). 
* No hardware patent is licensed hereunder.
*
* Redistributions must preserve existing copyright notices and reproduce this license (including the
* above copyright notice and the disclaimer and (if applicable) source code license limitations below)
* in the documentation and/or other materials provided with the distribution
*
* Redistribution and use in binary form, without modification, are permitted provided that the following
* conditions are met:
*
*   * No reverse engineering, decompilation, or disassembly of this software is permitted with respect to any
*     software provided in binary form.
*   * any redistribution and use are licensed by TI for use only with TI Devices.
*   * Nothing shall obligate TI to provide you with source code for the software licensed and provided to you in object code.
*
* If software source code is provided to you, modification and redistribution of the source code are permitted
* provided that the following conditions are met:
*
*   * any redistribution and use of the source code, including any resulting derivative works, are licensed by
*     TI for use only with TI Devices.
*   * any redistribution and use of any object code compiled from the source code and any resulting derivative
*     works, are licensed by TI for use only with TI Devices.
*
* Neither the name of Texas Instruments Incorporated nor the names of its suppliers may be used to endorse or
* promote products derived from this software without specific prior written permission.
*
* DISCLAIMER.
*
* THIS SOFTWARE IS PROVIDED BY TI AND TI’S LICENSORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
* BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL TI AND TI’S LICENSORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
* OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/

module deser_cmos(
   
    input CMOS_DIN_A,                   //input serial data D0 (CMOS logic)
    input CMOS_DIN_B,                   //input serial data D1 (CMOS logic)
    input CMOS_DIN_C,                   //input serial data D2 (CMOS logic)
    input CMOS_DIN_D,                   //input serial data D3 (CMOS logic)
    input DCLK,                         //input DCLK (CMOS logic)
    input FCLK,                         //input FCLK (CMOS logic)
           
    input data_rate,                    //DDR or SDR selection (DDR : data_rate = 0; SDR : data_rate = 1)
    input [1:0] data_lanes,             //no:of lanes;  0 = 1 lane; 1 = 2 lane, 2 = 4 lane 
    
    input RST,
    output DRDY,                        //data ready signal (indicates the final deserialized output)
    output [191:0] DOUT                 //deserialized output
    
    );


//iDDR instantiation
    
wire pos_DIN_A, neg_DIN_A;

iDDR iDDR_DIN_A (
    .DIN (CMOS_DIN_A),
    .CK (DCLK),
    .Q1 (pos_DIN_A),
    .Q2 (neg_DIN_A),
    .CE (1)
);

wire pos_DIN_B, neg_DIN_B;

iDDR iDDR_DIN_B (
    .DIN (CMOS_DIN_B),
    .CK (DCLK),
    .Q1 (pos_DIN_B),
    .Q2 (neg_DIN_B),
    .CE (1)
);

wire pos_DIN_C, neg_DIN_C;

iDDR iDDR_DIN_C (
    .DIN (CMOS_DIN_C),
    .CK (DCLK),
    .Q1 (pos_DIN_C),
    .Q2 (neg_DIN_C),
    .CE (1)
);

wire pos_DIN_D, neg_DIN_D;

iDDR iDDR_DIN_D (
    .DIN (CMOS_DIN_D),
    .CK (DCLK),
    .Q1 (pos_DIN_D),
    .Q2 (neg_DIN_D),
    .CE (1)
);


   
wire pos_CMOS_FCLK, neg_CMOS_FCLK;

iDDR iDDR_FCLK (
    .DIN (FCLK),                
    .CK (DCLK),
    .Q1 (pos_CMOS_FCLK),
    .Q2 (neg_CMOS_FCLK),
    .CE (1)
);


    
// SDR DESERIALIZER --------------------------------------------------------------------

reg [191:0] shift_reg_d;                    //D3
reg [95:0] shift_reg_b;                     //D1
reg [47:0] shift_reg_a, shift_reg_c;        //D0 and D2
wire [191:0] dout_sdr_1lane, dout_sdr_2lane, dout_sdr_4lane;

//192 bits are aligned as {ADC A {191:96},  ADC B {95:0}}
always @(negedge DCLK) begin                //latching at the neg edge of DCLK

    shift_reg_d[191:1] <= shift_reg_d[190:0];
    shift_reg_d[0] <= CMOS_DIN_D;
    
    shift_reg_b[95:1] <= shift_reg_b[94:0];
    shift_reg_b[0] <= CMOS_DIN_B;

    shift_reg_c[47:1] <= shift_reg_c[46:0];
    shift_reg_c[0] <= CMOS_DIN_C;

    shift_reg_a[47:1] <= shift_reg_a[46:0];
    shift_reg_a[0] <= CMOS_DIN_A;
end

    assign dout_sdr_1lane = shift_reg_d;                    //D3 has all the data
    
genvar i;
generate
    
    for(i = 0; i < 96; i = i + 1) begin
        assign dout_sdr_2lane[i] = shift_reg_b[i];          //D3 all bits of ADC A
        assign dout_sdr_2lane[96 + i] = shift_reg_d[i];     //D1 all bits of ADC B
    end
    
    for(i = 0; i < 48; i = i + 1) begin
        assign dout_sdr_4lane[2*i] = shift_reg_a[i];        //D2 LSB of ADC A
        assign dout_sdr_4lane[2*i + 1] = shift_reg_b[i];    //D3 MSB of ADC A
        assign dout_sdr_4lane[96 + 2*i] = shift_reg_c[i];   //D0 LSB of ADC B
        assign dout_sdr_4lane[97 + 2*i] = shift_reg_d[i];   //D1 MSB of ADC B
    end
endgenerate

// DDR DESERIALIZER ----------------------------------------------------------------

reg [96:0] pos_shift_reg_d, neg_shift_reg_d;                
reg [48:0] pos_shift_reg_b, neg_shift_reg_b;                
reg [24:0] pos_shift_reg_c, neg_shift_reg_c;                
reg [24:0] pos_shift_reg_a, neg_shift_reg_a;                
wire [191:0] dout_ddr_1lane, dout_ddr_1lane_neg, dout_ddr_2lane, dout_ddr_2lane_neg, dout_ddr_4lane, dout_ddr_4lane_neg;

always @(posedge DCLK) begin

    pos_shift_reg_d[96:1] <= pos_shift_reg_d[95:0];
    pos_shift_reg_d[0] <= pos_DIN_D;
    neg_shift_reg_d[96:1] <= neg_shift_reg_d[95:0];
    neg_shift_reg_d[0] <= neg_DIN_D;

    pos_shift_reg_b[48:1] <= pos_shift_reg_b[47:0];
    pos_shift_reg_b[0] <= pos_DIN_B;
    neg_shift_reg_b[48:1] <= neg_shift_reg_b[47:0];
    neg_shift_reg_b[0] <= neg_DIN_B;
    
    pos_shift_reg_c[24:1] <= pos_shift_reg_c[23:0];
    pos_shift_reg_c[0] <= pos_DIN_C;
    neg_shift_reg_c[24:1] <= neg_shift_reg_c[23:0];
    neg_shift_reg_c[0] <= neg_DIN_C;
    
    pos_shift_reg_a[24:1] <= pos_shift_reg_a[23:0];
    pos_shift_reg_a[0] <= pos_DIN_A;
    neg_shift_reg_a[24:1] <= neg_shift_reg_a[23:0];
    neg_shift_reg_a[0] <= neg_DIN_A;
end


//192 bits are aligned as {ADC A {191:96},  ADC B {95:0}}
generate
    for(i = 0; i < 96; i = i + 1) begin
        assign dout_ddr_1lane[2*i] = pos_shift_reg_d[i];                
        assign dout_ddr_1lane[2*i + 1] = neg_shift_reg_d[i];            
    end
    
    for(i = 0; i < 96; i = i + 1) begin
        assign dout_ddr_1lane_neg[2*i] = neg_shift_reg_d[i];             
        assign dout_ddr_1lane_neg[2*i + 1] = pos_shift_reg_d[i+1];       
    end
    
    for(i = 0; i < 48; i = i + 1) begin
        assign dout_ddr_2lane[2*i] = pos_shift_reg_b[i];                
        assign dout_ddr_2lane[2*i + 1] = neg_shift_reg_b[i];            
        assign dout_ddr_2lane[96 + 2*i] = pos_shift_reg_d[i];           
        assign dout_ddr_2lane[97 + 2*i] = neg_shift_reg_d[i];           
    end
    
    for(i = 0; i < 48; i = i + 1) begin
        assign dout_ddr_2lane_neg[2*i] = neg_shift_reg_b[i];             
        assign dout_ddr_2lane_neg[2*i + 1] = pos_shift_reg_b[i+1];       
        assign dout_ddr_2lane_neg[96 + 2*i] = neg_shift_reg_d[i];        
        assign dout_ddr_2lane_neg[97 + 2*i] = pos_shift_reg_d[i+1];      
    end
    
    for(i = 0; i < 24; i = i + 1) begin
        assign dout_ddr_4lane[4*i] = pos_shift_reg_a[i];                
        assign dout_ddr_4lane[4*i + 1] = pos_shift_reg_b[i];            
        assign dout_ddr_4lane[4*i + 2] = neg_shift_reg_a[i];            
        assign dout_ddr_4lane[4*i + 3] = neg_shift_reg_b[i];            
        assign dout_ddr_4lane[96 + 4*i] = pos_shift_reg_c[i];           
        assign dout_ddr_4lane[97 + 4*i] = pos_shift_reg_d[i];           
        assign dout_ddr_4lane[98 + 4*i] = neg_shift_reg_c[i];           
        assign dout_ddr_4lane[99 + 4*i] = neg_shift_reg_d[i];           
    end
    
    for(i = 0; i < 24; i = i + 1) begin
        assign dout_ddr_4lane_neg[4*i] = neg_shift_reg_a[i];            
        assign dout_ddr_4lane_neg[4*i + 1] = neg_shift_reg_b[i];        
        assign dout_ddr_4lane_neg[4*i + 2] = pos_shift_reg_a[i+1];      
        assign dout_ddr_4lane_neg[4*i + 3] = pos_shift_reg_b[i+1];      
        assign dout_ddr_4lane_neg[96 + 4*i] = neg_shift_reg_c[i];       
        assign dout_ddr_4lane_neg[97 + 4*i] = neg_shift_reg_d[i];       
        assign dout_ddr_4lane_neg[98 + 4*i] = pos_shift_reg_c[i+1];     
        assign dout_ddr_4lane_neg[99 + 4*i] = pos_shift_reg_d[i+1];     
    end
endgenerate


// DATA OUTPUT MUX -----------------------------------------------------------------------------------------------
//if DDR (data_rate = 0) then data is interleaved between pos_shift_reg [95:0] and neg_shift_reg [95:0]
//if SDR (data_rate = 1) then all data is in pos_shift_reg [191:0]

reg FCLK_pos;

//final deserialized output selection

assign DOUT =   (data_rate == 0 && data_lanes == 2'b00 && FCLK_pos == 0)?(dout_ddr_1lane):
                (data_rate == 0 && data_lanes == 2'b00 && FCLK_pos == 1)?(dout_ddr_1lane_neg): 
                (data_rate == 0 && data_lanes == 2'b01 && FCLK_pos == 0)?(dout_ddr_2lane):
                (data_rate == 0 && data_lanes == 2'b01 && FCLK_pos == 1)?(dout_ddr_2lane_neg):     
                (data_rate == 0 && data_lanes == 2'b10 && FCLK_pos == 0)?(dout_ddr_4lane):
                (data_rate == 0 && data_lanes == 2'b10 && FCLK_pos == 1)?(dout_ddr_4lane_neg):
                (data_rate == 1 && data_lanes == 2'b00)?(dout_sdr_1lane):
                (data_rate == 1 && data_lanes == 2'b01)?(dout_sdr_2lane):
                (data_rate == 1 && data_lanes == 2'b10)?(dout_sdr_4lane):(192'hBAD00BAD00BAD00BAD00BAD00BAD00BAD00BAD00BAD00BAD);


//generation of DRDY signal
//When FCLK_pos = HIGH, mux the dout_ddr_4lane_neg; when FCLK_pos = LOW, mus the dout_ddr_4lane

reg fclk_ddr_dly1;
reg fclk_ddr_dly2;
reg fclk_ddr_dly3;

always @(negedge DCLK) begin
    fclk_ddr_dly1 <= pos_CMOS_FCLK | neg_CMOS_FCLK;
    fclk_ddr_dly2 <= fclk_ddr_dly1;
    fclk_ddr_dly3 <= fclk_ddr_dly2;
end


wire fclk_posedge;
posedge_detect FCLK_rising_edge(
    .RST(RST),
    .sig (pos_CMOS_FCLK & ~neg_CMOS_FCLK),
    .clk (DCLK),
    .pe (fclk_posedge)
    );

    
always @(posedge DCLK) begin
    if( RST == 1 )
        FCLK_pos <= 1'b0;
    else
        FCLK_pos <= FCLK_pos | fclk_posedge;
end

assign DRDY = ((FCLK_pos == 0)&& (data_rate == 0)) ? (fclk_ddr_dly1) :
              ((FCLK_pos == 1) && (data_rate == 0)) ? (fclk_ddr_dly2):
              (data_rate == 1) ?  (pos_CMOS_FCLK | neg_CMOS_FCLK): 
              1'b0;
              
endmodule



module posedge_detect(
    input RST,
    input sig,
    input clk,
    output pe
    );
    

reg sig_dly1;
reg sig_dly2;
reg sig_dly3;

always @ (posedge clk)
begin
    if(RST == 1)    begin
        sig_dly1 <= 0;
        sig_dly2 <= 0;
        sig_dly3 <= 0;
    end
    
    else    begin
        sig_dly1 <= sig;
        sig_dly2 <= sig_dly1;
        sig_dly3 <= sig_dly2;
    end
end

assign pe = sig_dly2 & ~sig_dly3;

endmodule

